/**
  *  \file client/wcost.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wcost.h"
#include "game/cargo.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "util/misc.h"
#include "game/planet.h"
#include "game/actions/cargocost.h"

WItemCostDisplay::WItemCostDisplay(const GPlanet& planet, bool include_item_in_cost)
    : planet(planet), part_cost(), total_cost(), have_tech(1), need_tech(1),
      include_item_in_cost(include_item_in_cost)
{
    setState(st_Disabled, false);
}

WItemCostDisplay::~WItemCostDisplay()
{ }

void
WItemCostDisplay::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: should explicitly compute table size
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(350, 7*font_heights[FONT_NORMAL]);
}

bool
WItemCostDisplay::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIWidget::handleEvent(e, second_pass);
}

void
WItemCostDisplay::writeCost(UIAbstractTableWriter& tw, int32 need, int32 remain, int32 item)
{
    if (remain < 0 || (include_item_in_cost && remain < item))
        tw.writeVariableCell(UIColor::tc_Red, numToString(need));
    else
        tw.writeVariableCell(UIColor::tc_Green, numToString(need));
}

void
WItemCostDisplay::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Tritanium:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Duranium:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Molybdenum:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Megacredits:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tech Level:"));

    // "This part". Color is red if resources don't suffice to buy one of these.
    // With include_item_in_cost, color is red if resources don't suffice to buy one more.
    tw.newRColumn(5 * font_ems[FONT_NORMAL]);
    tw.writeFixedCell(UIColor::tc_Contrast, _("This Part"));
    writeCost(tw, part_cost.get(el_Tritanium),    planet.getCargoRaw(el_Tritanium)   - part_cost.get(el_Tritanium),  total_cost.get(el_Tritanium));
    writeCost(tw, part_cost.get(el_Duranium),     planet.getCargoRaw(el_Duranium)    - part_cost.get(el_Duranium),   total_cost.get(el_Duranium));
    writeCost(tw, part_cost.get(el_Molybdenum),   planet.getCargoRaw(el_Molybdenum)  - part_cost.get(el_Molybdenum), total_cost.get(el_Molybdenum));
    writeCost(tw, part_cost.get(el_Money),        planet.getCargoRaw(el_Money) + planet.getCargoRaw(el_Supplies) - part_cost.get(el_Supplies) - part_cost.get(el_Money), total_cost.get(el_Money) + total_cost.get(el_Supplies));
    if (part_cost.get(el_Supplies))
        writeCost(tw, part_cost.get(el_Supplies), planet.getCargoRaw(el_Supplies)    - part_cost.get(el_Supplies),   total_cost.get(el_Supplies));
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    writeCost(tw, need_tech, have_tech - need_tech, 0);

    // "Total"
    tw.newRColumn(5 * font_ems[FONT_NORMAL]);
    tw.writeFixedCell(UIColor::tc_Contrast, _("Total"));
    writeCost(tw, total_cost.get(el_Tritanium),    planet.getCargoRaw(el_Tritanium)   - total_cost.get(el_Tritanium), 0);
    writeCost(tw, total_cost.get(el_Duranium),     planet.getCargoRaw(el_Duranium)    - total_cost.get(el_Duranium), 0);
    writeCost(tw, total_cost.get(el_Molybdenum),   planet.getCargoRaw(el_Molybdenum)  - total_cost.get(el_Molybdenum), 0);
    writeCost(tw, total_cost.get(el_Money),        planet.getCargoRaw(el_Money) + planet.getCargoRaw(el_Supplies) - total_cost.get(el_Supplies) - total_cost.get(el_Money), 0);
    if (total_cost.get(el_Supplies))
        writeCost(tw, total_cost.get(el_Supplies), planet.getCargoRaw(el_Supplies)    - total_cost.get(el_Supplies), 0);
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());

    // Available
    tw.newRColumn(5 * font_ems[FONT_NORMAL]);
    tw.writeFixedCell(UIColor::tc_Contrast, _("Available"));
    writeCost(tw, planet.getCargoRaw(el_Tritanium),  0, 0);
    writeCost(tw, planet.getCargoRaw(el_Duranium),   0, 0);
    writeCost(tw, planet.getCargoRaw(el_Molybdenum), 0, 0);
    writeCost(tw, planet.getCargoRaw(el_Money),      0, 0);
    writeCost(tw, planet.getCargoRaw(el_Supplies),   0, 0);
    writeCost(tw, have_tech, 0, 0);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "mc");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, string_t());

    tw.finish();

    can.update(getExtent());
}

void
WItemCostDisplay::setPartCost(const GCost& it)
{
    part_cost = it;
    drawWidget(false);
}

void
WItemCostDisplay::setPartTech(int have, int need)
{
    have_tech = have;
    need_tech = need;
    drawWidget(false);
}

void
WItemCostDisplay::setTotalCost(const GCost& it)
{
    total_cost = it;
    drawWidget(false);
}




WObservingItemCostDisplay::WObservingItemCostDisplay(GCargoCostTransaction& action,
                                                     const GPlanet& planet,
                                                     bool include_item_in_cost)
    : WItemCostDisplay(planet, include_item_in_cost),
      conn_cost_change(action.sig_cost_changed.add(this, &WObservingItemCostDisplay::onCostChange)),
      action(action)
{ }

WObservingItemCostDisplay::~WObservingItemCostDisplay()
{ }

void
WObservingItemCostDisplay::onCostChange()
{
    setTotalCost(action.getCost());
}
