/**
  *  \file gfx/gfxxor.cc
  *  \brief "XOR" mode drawing.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/gfx.h"
#include "cpluslib/assert.h"

/** \class GfxXorFilter
    \brief A canvas that supports "XOR" drawing.

    Instead of putting this into the "core" code and passing around
    the "XOR" status explicitly all the time, we use a separate class.
    This one "peeks" into the underlying canvas, and uses explicit
    draw commands to do actual drawing.

    Pros: simple. Zero overhead for non-XOR. Output clipping is natural.

    Cons: slow. XOR drawing is intended for small markers only, anyway.

    The semantics of transparency for XOR drawing is a bit weird
    currently; the XOR results are drawn in that alpha level (losing
    us the "draw again to restore original" property). blitSurface and
    blitPattern are not supported. */

GfxXorFilter::GfxXorFilter(GfxCanvas& parent)
    : parent(parent)
{ }

GfxXorFilter::~GfxXorFilter()
{ }

void
GfxXorFilter::drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxPoint pt2 = pt;
    GfxPoint psize = parent.getSize();
    if (pt.x >= psize.x || pt.x + npix < 0 || pt.y < 0 || pt.y >= psize.y || pat == 0)
        return;

    while (npix > 0) {
        if (pat & (0x80 >> (pt2.x & 7)))
            parent.drawPixel(pt2, parent.getPixel(pt2) ^ color, alpha);
        ++pt2.x;
        --npix;
    }
}

void
GfxXorFilter::drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxPoint pt2 = pt;
    GfxPoint psize = parent.getSize();
    if (pt.x < 0 || pt.x >= psize.x || pt.y >= psize.y || pt.y + npix < 0 || pat == 0)
        return;

    while (npix > 0) {
        if (pat & (0x80 >> (pt2.y & 7)))
            parent.drawPixel(pt2, parent.getPixel(pt2) ^ color, alpha);
        ++pt2.y;
        --npix;
    }
}

void
GfxXorFilter::drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha)
{
    parent.drawPixel(pt, parent.getPixel(pt) ^ color, alpha);
}

void
GfxXorFilter::drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha)
{
    if (rect.w > 0) {
        while (rect.h > 0) {
            drawHLine(GfxPoint(rect.x, rect.y), rect.w, color, pat[rect.y & 7], alpha);
            if (bg != GFX_TRANSPARENT)
                drawHLine(GfxPoint(rect.x, rect.y), rect.w, bg, ~pat[rect.y & 7], alpha);
            --rect.h;
            ++rect.y;
        }
    }
}

void
GfxXorFilter::blitSurface(const GfxPoint& /*pt*/, GfxCanvas& /*src*/, GfxRect /*rect*/)
{
    FAIL();
}

void
GfxXorFilter::blitPattern(GfxRect /*rect*/, const GfxPoint& /*pt*/, int /*bytes_per_line*/, const uint8* /*data*/, uint32 /*color*/, uint32 /*bg*/, alpha_type /*alpha*/)
{
    // FIXME: this one should probably be implemented
    FAIL();
}

bool
GfxXorFilter::lock()
{
    return parent.lock();
}

void
GfxXorFilter::unlock()
{
    parent.unlock();
}

void
GfxXorFilter::update(GfxRect r)
{
    parent.update(r);
}

GfxRect
GfxXorFilter::computeClipRect(GfxRect r)
{
    return parent.computeClipRect(r);
}

GfxPoint
GfxXorFilter::getSize()
{
    return parent.getSize();
}

int
GfxXorFilter::getBitsPerPixel()
{
    return parent.getBitsPerPixel();
}

uint32
GfxXorFilter::getPixel(const GfxPoint& pt)
{
    return parent.getPixel(pt);
}
