/**
  *  \file reshack/colorsel.cc
  *  \brief Resource Editor: Color Selection for Pictures
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/colorsel.h"
#include "reshack/painter.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "reshack/palette.h"

/**************************** RHColorSelector ****************************/

/** Constructor.
    \param painter Associated painter */
RHColorSelector::RHColorSelector(RHPainter& painter)
    : painter(painter),
      page(0),
      mouse_down(false)
{
    observe(painter.sig_color_change);
}

RHColorSelector::~RHColorSelector()
{ }

/** "Next page" command. */
void
RHColorSelector::onNextPage()
{
    page = (page+1) & 15;
    drawWidget();
}

/** "Previous page" command. */
void
RHColorSelector::onPreviousPage()
{
    page = (page+15) & 15;
    drawWidget();
}

/** "Edit color" command. */
void
RHColorSelector::editColor(uint8_t color)
{
    RHPaletteEditorWindow win(painter.getPixmap(), color);
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}

void
RHColorSelector::drawContent(GfxCanvas& can)
{
    // 16 lines:
    //   nnn M[cc]N Pal
    //    |  |  | |  '- "can edit palette" indicator
    //    |  |  | `- "background color" marker
    //    |  |  `- color box
    //    |  `- "foreground color" marker
    //    `- color number
    // 17th line:
    //   [<] 0xNN [>]
    //    |   |    `- "next page"
    //    |   `- base color
    //    `- "previous page"
    // => 17x12 characters

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_FIXED]);
    ctx.useColorScheme(getSkin());

    // Draw textual/skinnable items
    drawBackground(ctx, extent);
    ctx.setColor(UIColor::tc_Static);
    for (int i = 0; i < 16; ++i) {
        uint8_t color = i + 16*page;
        int y = extent.y + font_heights[FONT_FIXED]*i;
        outText(ctx, extent.x, y, format("%3d", color));
        if (isEditableColor(painter.getColorMode(), color)) {
            outText(ctx, extent.x + 9*font_ems[FONT_FIXED], y, _("Pal"));
        }
        if (color == painter.getColor(false)) {
            outText(ctx, extent.x + 4*font_ems[FONT_FIXED], y, ">");
        }
        if (color == painter.getColor(true)) {
            outText(ctx, extent.x + 7*font_ems[FONT_FIXED], y, "<");
        }
    }
    outText(ctx, extent.x, extent.y + 16*font_heights[FONT_FIXED], format("[<] 0x%X0 [>]", page));

    // Draw color boxes
    ctx.useColorScheme(GfxNullColorScheme::instance);
    for (int i = 0; i < 16; ++i) {
        uint8_t color = i + 16*page;
        drawSolidBar(ctx, GfxRect(extent.x + 5*font_ems[FONT_FIXED], extent.y + font_heights[FONT_FIXED]*i,
                                  2*font_ems[FONT_FIXED], font_heights[FONT_FIXED]),
                     colorIndexToScreenColor(painter.getPixmap(), color));
    }

    can.update(getExtent());
}

bool
RHColorSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == SDLK_RIGHT + ss_Ctrl) {
            setSelectedWidget();
            onNextPage();
            return true;
        }
        if (e.key.code == SDLK_LEFT + ss_Ctrl) {
            setSelectedWidget();
            onPreviousPage();
            return true;
        }
    }
    if (e.type == ev_Mouse) {
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
            // Mouse is inside this widget, so consume this event
            int x = (e.mouse.x - extent.x) / font_ems[FONT_FIXED];
            int y = (e.mouse.y - extent.y) / font_heights[FONT_FIXED];
            uint32_t color = 16*page + y;
            if (e.mouse.b & mb_Left) {
                // Left mouse button pressed
                mouse_down = true;
                setSelectedWidget();
                if (x < 9 && y < 16 && color != painter.getColor(0)) {
                    painter.setColor(0, color);
                    drawWidget();
                } else if (x >= 9 && y < 16 && isEditableColor(painter.getColorMode(), color)) {
                    editColor(color);
                    drawWidget();
                    painter.drawWidget();
                }
                return true;
            } else if (e.mouse.b & mb_Right) {
                // Right mouse button pressed
                mouse_down = true;
                setSelectedWidget();
                if (x < 9 && y < 16 && color != painter.getColor(1)) {
                    painter.setColor(1, color);
                    drawWidget();
                }
                return true;
            } else if (mouse_down) {
                // Mouse button released. Check location.
                mouse_down = false;
                setSelectedWidget();
                if (x >= 9 && y < 16 && isEditableColor(painter.getColorMode(), color)) {
                    // FIXME: palette editor
                } else if (y == 16) {
                    if (x < 3) {
                        onPreviousPage();
                        return true;
                    } else if (x >= 9) {
                        onNextPage();
                        return true;
                    } else {
                        // FIXME: 16x16 selector
                    }
                }
            }
        } else if (mouse_down) {
            // Mouse is has left this widget
            mouse_down = false;
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
RHColorSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = font_sizes[FONT_FIXED].scaledBy(12, 17);
}

void
RHColorSelector::onStateChange(int bits, bool enable)
{
    if ((bits & st_Selected) != 0 && !enable)
        mouse_down = false;
    UIWidget::onStateChange(bits, enable);
}

/************************** RHFontColorSelector **************************/

enum { NFONTCOLORS = 3 };

static const uint8_t font_colors[NFONTCOLORS] = { FC_BLACK, FC_HALF, FC_WHITE };

/** Constructor.
    \param painter Associated painter */
RHFontColorSelector::RHFontColorSelector(RHPainter& painter)
    : painter(painter)
{
    observe(painter.sig_color_change);
}

RHFontColorSelector::~RHFontColorSelector()
{ }

void
RHFontColorSelector::drawContent(GfxCanvas& can)
{
    // >[nnnn]<
    // |  |   `- background color indicator
    // |  `- color block
    // `- foreground color indicator

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_FIXED]);
    ctx.useColorScheme(getSkin());

    // Draw color indicators
    drawBackground(ctx, extent);
    ctx.setColor(UIColor::tc_Static);
    for (int i = 0; i < NFONTCOLORS; ++i) {
        int y = extent.y + font_heights[FONT_FIXED]*i;
        uint32_t color = font_colors[i];
        if (color == painter.getColor(false)) {
            outText(ctx, extent.x, y, ">");
        }
        if (color == painter.getColor(true)) {
            outText(ctx, extent.x + extent.w - font_ems[FONT_FIXED], y, "<");
        }
    }

    // Draw color boxes
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    for (int i = 0; i < NFONTCOLORS; ++i) {
        drawSolidBar(ctx, GfxRect(extent.x + font_ems[FONT_FIXED], extent.y + font_heights[FONT_FIXED]*i,
                                  extent.w - 2*font_ems[FONT_FIXED], font_heights[FONT_FIXED]),
                     font_colors[i]);
    }

    can.update(getExtent());
}

bool
RHFontColorSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
            // Mouse is inside this widget, so consume this event
            int y = (e.mouse.y - extent.y) / font_heights[FONT_FIXED];
            if (e.mouse.b & mb_Left) {
                // Left mouse button pressed
                setSelectedWidget();
                if (y < NFONTCOLORS && font_colors[y] != painter.getColor(0)) {
                    painter.setColor(0, font_colors[y]);
                    drawWidget();
                }
                return true;
            } else if (e.mouse.b & mb_Right) {
                // Right mouse button pressed
                setSelectedWidget();
                if (y < NFONTCOLORS && font_colors[y] != painter.getColor(1)) {
                    painter.setColor(1, font_colors[y]);
                    drawWidget();
                }
                return true;
            }
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
RHFontColorSelector::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_FIXED].scaledBy(5, NFONTCOLORS);
    info.flags = 0;
}
