/**
  *  \file u/t_util_xmlread.cc
  *  \brief Unit test for util/xmlread
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "util/xmlread.h"

#include "u/t_util.h"
#include "io/memstream.h"

static const char theDocument[] = ("<?xml version='1.0'?>\n"
                                   "\n"
                                   "<foo>\n"
                                   "  < bar zot >\n"
                                   "  < / bar >\n"
                                   "  <text id=1>\x80</text>\n"
                                   "</foo>\n");

static void
testBasicDocument(XmlReader& xr, const char* name)
{
    XmlReader::Token t;

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::PIStart);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "xml");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::PIAttribute);
    TSM_ASSERT_EQUALS(name, xr.getName(), "version");
    TSM_ASSERT_EQUALS(name, xr.getValue(), "1.0");
    TSM_ASSERT_EQUALS(name, xr.getTag(), "xml");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagStart);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "foo");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagStart);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "bar");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagAttribute);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "bar");
    TSM_ASSERT_EQUALS(name, xr.getName(), "zot");
    TSM_ASSERT_EQUALS(name, xr.getValue(), "zot");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagEnd);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "bar");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagStart);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "text");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagAttribute);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "text");
    TSM_ASSERT_EQUALS(name, xr.getName(), "id");
    TSM_ASSERT_EQUALS(name, xr.getValue(), "1");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::Text);
    TSM_ASSERT_EQUALS(name, xr.getValue(), "\xc2\x80");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagEnd);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "text");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::TagEnd);
    TSM_ASSERT_EQUALS(name, xr.getTag(), "foo");

    t = xr.read();
    TSM_ASSERT_EQUALS(name, t, XmlReader::Eof);
}

void
UtilXmlreadTestSuite::testUtf8()
{
    {
        MemoryStream ms("<?xml version='1.0'?>\n"
                        "\n"
                        "<foo>\n"
                        "  < bar zot >\n"
                        "  < / bar >\n"
                        "  <text id=1>\xc2\x80</text>\n"
                        "</foo>\n");
        XmlReader xr(ms);
        testBasicDocument(xr, "testUtf8 (implicit)");
    }

    {
        MemoryStream ms("<?xml version='1.0' encoding=\"utf-8\"?>\n"
                        "\n"
                        "<text id=1>\xc2\x80</text>\n");
        XmlReader xr(ms);
        XmlReader::Token t;

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::PIStart);
        TS_ASSERT_EQUALS(xr.getTag(), "xml");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
        TS_ASSERT_EQUALS(xr.getName(), "version");
        TS_ASSERT_EQUALS(xr.getValue(), "1.0");
        TS_ASSERT_EQUALS(xr.getTag(), "xml");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
        TS_ASSERT_EQUALS(xr.getName(), "encoding");
        TS_ASSERT_EQUALS(xr.getValue(), "utf-8");
        TS_ASSERT_EQUALS(xr.getTag(), "xml");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::TagStart);
        TS_ASSERT_EQUALS(xr.getTag(), "text");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::TagAttribute);
        TS_ASSERT_EQUALS(xr.getTag(), "text");
        TS_ASSERT_EQUALS(xr.getName(), "id");
        TS_ASSERT_EQUALS(xr.getValue(), "1");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::Text);
        TS_ASSERT_EQUALS(xr.getValue(), "\xc2\x80");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
        TS_ASSERT_EQUALS(xr.getTag(), "text");

        t = xr.read();
        TS_ASSERT_EQUALS(t, XmlReader::Eof);
    }
}

void
UtilXmlreadTestSuite::testLatin1()
{
    MemoryStream ms("<?xml version='1.0' encoding='iso-8859-1'?>\n"
                    "\n"
                    "<foo>\n"
                    "  < bar zot >\n"
                    "  < / bar >\n"
                    "  <text id=1>\xc2\x80</text>\n"
                    "</foo>\n");
    XmlReader xr(ms);
    XmlReader::Token t;

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIStart);
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
    TS_ASSERT_EQUALS(xr.getName(), "version");
    TS_ASSERT_EQUALS(xr.getValue(), "1.0");
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
    TS_ASSERT_EQUALS(xr.getName(), "encoding");
    TS_ASSERT_EQUALS(xr.getValue(), "iso-8859-1");
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "foo");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagAttribute);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");
    TS_ASSERT_EQUALS(xr.getName(), "zot");
    TS_ASSERT_EQUALS(xr.getValue(), "zot");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "text");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagAttribute);
    TS_ASSERT_EQUALS(xr.getTag(), "text");
    TS_ASSERT_EQUALS(xr.getName(), "id");
    TS_ASSERT_EQUALS(xr.getValue(), "1");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::Text);
    TS_ASSERT_EQUALS(xr.getValue(), "\xc3\x82\xc2\x80");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "text");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "foo");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::Eof);
}

void
UtilXmlreadTestSuite::testKOI8R()
{
    MemoryStream ms("<?xml version='1.0' encoding='koi8r'?>\n"
                    "\n"
                    "<foo>\n"
                    "  < bar zot >\n"
                    "  < / bar >\n"
                    "  <text id=1>\xc2\x80</text>\n"
                    "</foo>\n");
    XmlReader xr(ms);
    XmlReader::Token t;

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIStart);
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
    TS_ASSERT_EQUALS(xr.getName(), "version");
    TS_ASSERT_EQUALS(xr.getValue(), "1.0");
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::PIAttribute);
    TS_ASSERT_EQUALS(xr.getName(), "encoding");
    TS_ASSERT_EQUALS(xr.getValue(), "koi8r");
    TS_ASSERT_EQUALS(xr.getTag(), "xml");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "foo");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagAttribute);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");
    TS_ASSERT_EQUALS(xr.getName(), "zot");
    TS_ASSERT_EQUALS(xr.getValue(), "zot");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "bar");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagStart);
    TS_ASSERT_EQUALS(xr.getTag(), "text");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagAttribute);
    TS_ASSERT_EQUALS(xr.getTag(), "text");
    TS_ASSERT_EQUALS(xr.getName(), "id");
    TS_ASSERT_EQUALS(xr.getValue(), "1");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::Text);
    TS_ASSERT_EQUALS(xr.getValue(), "\xD0\xB1\xE2\x94\x80");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "text");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::TagEnd);
    TS_ASSERT_EQUALS(xr.getTag(), "foo");

    t = xr.read();
    TS_ASSERT_EQUALS(t, XmlReader::Eof);
}

void
UtilXmlreadTestSuite::testUtf16LE()
{
    Buffer<char> b;
    for (std::size_t i = 0; theDocument[i] != 0; ++i) {
        b.append(theDocument[i]);
        b.append(0);
    }
    MemoryStream ms(b);
    XmlReader xr(ms);
    testBasicDocument(xr, "testUtf16LE");
}

void
UtilXmlreadTestSuite::testUtf16BE()
{
    Buffer<char> b;
    for (std::size_t i = 0; theDocument[i] != 0; ++i) {
        b.append(0);
        b.append(theDocument[i]);
    }
    MemoryStream ms(b);
    XmlReader xr(ms);
    testBasicDocument(xr, "testUtf16BE");
}

void
UtilXmlreadTestSuite::testText()
{
    /* The XML reader discards text elements containing whitespace
       only. Since CDATA blocks are treated as separate elements, the
       newline between two CDATA blocks would be eaten. We therefore
       add a 'x'. */
    MemoryStream ms("a\n"
                    "<![CDATA[]]>x\n"
                    "<![CDATA[[]]]>x\n"
                    "<![CDATA[<foo>]]>x\n"
                    "<![CDATA[<?bar?>]]>x\n"
                    "<![CDATA[ &amp; ]]>x\n"
                    "<![CDATA[a]b]]]>x\n"
                    " &amp;\n"
                    " &quot;\n"
                    "b\n"
                    "&#65;&#x61;\n"
                    "&#160;\n"
                    "&#xa0;\n");
    XmlReader xr(ms);
    XmlReader::Token t;
    string_t result;
    while ((t = xr.read()) != XmlReader::Eof) {
        TS_ASSERT_EQUALS(t, XmlReader::Text);
        result += xr.getValue();
    }
    TS_ASSERT_EQUALS(result,
                     "a\n"
                     "x\n"
                     "[]x\n"
                     "<foo>x\n"
                     "<?bar?>x\n"
                     " &amp; x\n"
                     "a]b]x\n"
                     " &\n"
                     " \"\n"
                     "b\n"
                     "Aa\n"
                     "\xc2\xa0\n"
                     "\xc2\xa0\n");
}
